# Copyright 2019 Oleg Butuzov. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
deadlinks.crawler
~~~~~~~~~~~~~~~~~

Crawl the links on from the provided start point.

:copyright: (c) 2019 by Oleg Butuzov.
:license:   Apache2, see LICENSE for more details.
"""

# -- Imports -------------------------------------------------------------------

from typing import (List, Tuple, Optional, Dict)
from types import FrameType

from threading import Thread
from signal import (signal, SIGINT)
from queue import Queue
from time import sleep

from .link import Link
from .status import Status
from .index import Index
from .settings import Settings
from .exceptions import (
    DeadlinksIgnoredURL,
    DeadlinksRedirectionURL,
)

# -- Implementation ------------------------------------------------------------


class Crawler:
    """ Crawler/Spider Application. """

    def __init__(self, settings: Settings) -> None:
        """ Performe Crawling described by Settings. """

        self.settings = settings
        self.retry = settings.retry
        self.index = Index()

        # Application state
        self.terminated = False # type: bool
        self.crawling = False # type: bool
        self.crawled = False # type: bool
        self.queue = Queue() # type: Queue

        # Initialization of the Queue and Index
        self._base = settings.base

        # Checking if this is Ignored URL
        is_ignored, message = self.is_ignored(settings.base)
        if is_ignored:
            self._base.status = Status.IGNORED
            self._base.message = message
            error = "Issue with Base URL <{}>: {}"
            raise DeadlinksIgnoredURL(error.format(self._base.url(), message))

        self.add(self._base)

    def stop(self, sig: int, frame: FrameType) -> None:
        """ Captures SIGINT signal and and change terminition state """
        self.terminated = True

    def terminition_watcher(self) -> None:
        """ Clears queue on teminated execution

        TODO: Test long term waiting page and terminition
        """

        while not self.terminated:
            sleep(0.01)

        while True:
            try:
                while not self.queue.empty():
                    self.queue.task_done()
                sleep(0.01)
            except ValueError:
                break

    def start(self) -> None:
        """ Starts the crawling process. """

        if self.crawling or self.crawled or self.terminated:
            return

        self.crawling = True

        # catching kill signal.
        signal(SIGINT, self.stop)

        if self.settings.threads > 1:

            thread = Thread(target=self.terminition_watcher, daemon=True)
            thread.start()

            for idx in range(1, 1 + self.settings.threads):
                thread = Thread(target=self.indexer, args=[idx], daemon=True)
                thread.start()

            self.queue.join()
        else:
            self.indexer()

        self.crawling = False
        self.crawled = True

    def indexer(self, thread_number: int = 0) -> None:
        """ Indexation process. """

        while True:
            while not self.terminated and not self.queue.empty():

                url = self.queue.get()
                self.update(url)
                self.queue.task_done()

            else:

                # catching terminated
                if self.terminated:
                    return

                # in case is its only 1 thread we NOT waiting for
                # new url to come.
                if thread_number == 0:
                    break

                # and we wait for fraction of second, if there are many threads.
                sleep(thread_number / 10)

    def add(self, link: Link) -> None:
        """ Queue URL. """
        if link in self.index:
            return

        self.index.put(link)
        self.queue.put(link)

    def is_ignored(self, url: Link) -> Tuple[bool, Optional[str]]:
        """ Check if url can be ignored """

        # We can have few different cases when URL is ignored.

        # ok, checking for valid link first
        if not url.is_valid():
            return (True, "URL isn't valid")

        # TODO - Site Owner Ask (via meta tag) to ignore this URL
        # https://support.google.com/webmasters/answer/93710?hl=en

        # TODO - Site Owner Ask (via robots.txt) to ignore this URL
        # https://www.robotstxt.org/robotstxt.html

        # We have an external URL and cheking external URL are Off
        if url.is_external(self.settings.base) and not self.settings.external:
            return (True, "External indexation is Off")

        # This is a URL that fits to one of the ignored domains
        if url.match_domains(self.settings.domains):
            return (True, "Matching ignored domain")

        # This is a URL that fits to one of the ignored pathes
        if url.match_pathes(self.settings.pathes):
            return (True, "Matching ignored path")

        # FORBIDDEN: This is a local URL that located outside indexed path.
        if not url.is_external(self.settings.base) and  \
            self.settings.stay_within_path and not self._base.within(url):
            return (True, "URL outside of the allowed path")

        return (False, None)

    def update(self, url: Link) -> None:
        """ Update state or the url by checking it's data. """

        # We assume that URL Link that passed to this method is in status
        # Status.UNDEFINED, therefor we can perform required checks.
        if url in self.index:
            url = self.index[url]

        if url.status != Status.UNDEFINED:
            return

        # Adding URL to index so we can track its state.
        is_ignored, message = self.is_ignored(url)

        if is_ignored:
            self.index.update(url, Status.IGNORED, str(message))
            return

        try:
            # TODO - rething short calls implementation.
            is_external = url.is_external(self.settings.base)
            if not url.exists(is_external, retries=self.retry):
                self.index.update(url, Status.NOT_FOUND, url.message)
                return
        except DeadlinksRedirectionURL as _href:
            # ok, so next time we looking for this
            # we will need to make lookup to redirected URL.
            self.index.update(url, Status.REDIRECTION, "")
            self.add_and_go(url, str(_href))
            return
        except DeadlinksIgnoredURL:
            # we catching this exception jic, but "it should never happen"
            return

        # we defining status of this url as FOUND
        self.index.update(url, Status.FOUND, "")

        for href in url.links:
            self.add_and_go(url, href)

    def add_and_go(self, url: Link, href: str) -> None:
        """ Reducing code duplication. """

        # Create link variable of type Link that represent a new
        #    relative to URL link that has href
        link = Link(url.link(href)) # type: Link

        # adding it to queue
        self.add(link)

        # Update link by adding url as link referrer
        link.add_referrer(url.url())

    def internal(self, links: List[Link]) -> List[Link]:
        """ Masks URL for local files, so it looks like internal domain real. """

        if not self.settings.masked:
            return links

        domain = self.settings.base.domain
        mapper = lambda x: x.url().replace(domain, 'internal')

        return list(map(Link, map(mapper, links)))

    @property
    def ignored(self) -> List[Link]:
        """ Return URLs we have ignore to check. """
        return self.internal(self.index.ignored())

    @property
    def succeed(self) -> List[Link]:
        """ Return URLs we exists. """
        return self.internal(self.index.succeed())

    @property
    def failed(self) -> List[Link]:
        """ Return URLs failed to exist. """
        return self.internal(self.index.failed())

    @property
    def redirected(self) -> List[Link]:
        """ Return URLs failed to exist. """
        return self.internal(self.index.redirected())

    @property
    def stats(self) -> Dict[Status, int]:
        """ return crawler stats """
        return self.index.get_stats()



# Copyright 2019 Oleg Butuzov. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
deadlinks.settings
~~~~~~~~~~~~~~~~~~

Handles settings passed to crawler.

:copyright: (c) 2019 by Oleg Butuzov.
:license:   Apache2, see LICENSE for more details.
"""

# -- Imports -------------------------------------------------------------------

from typing import (Optional, List, Dict, Any, Union)

from pathlib import Path

from .baseurl import BaseURL
from .serving import Server

from .exceptions import (
    DeadlinksSettingsThreads,
    DeadlinksSettingsBase,
    DeadlinksSettingsRetry,
    DeadlinksSettingsChange,
    DeadlinksSettingsDomains,
    DeadlinksSettingsPathes,
    DeadlinksSettingsPath,
    DeadlinksSettingsRoot,
)

# -- Implementation ------------------------------------------------------------
# TODO - Review, and may be somehow simplify arguments


class Settings:
    """ Handles general settings. """
    # pylint: disable=R0902
    _stay_within_path = None # type: Optional[bool]
    _external = None # type: Optional[bool]
    _threads = None # type: Optional[int]
    _domains = None # type: Optional[List[str]]
    _pathes = None # type: Optional[List[str]]
    _retry = None # type: Optional[int]
    _base = None # type: Optional[BaseURL]
    _root = None # type: Optional[Path]

    def __init__(self, url: str, **kwargs: Any) -> None:
        """ Instantiate settings class. """

        defaults = Settings.defaults(kwargs)

        # first we adding misc properties
        self.threads = defaults['threads']
        self.external = defaults['check_external_urls']
        self.retry = defaults['retry']

        # ignoring pathes.
        self.domains = defaults['ignore_domains']
        self.pathes = defaults['ignore_pathes']

        self.stay_within_path = defaults['stay_within_path']

        # next we create base url and check for ignore patterns
        self._is_masked = False
        # special case of baseurl is internal
        base = BaseURL(url)
        if base.domain == "internal":
            self._is_masked = True
            self.root = Path(defaults['root']) # type: ignore
            web_server = Server(self.root, base.path)
            base = BaseURL(web_server.url())

        self.base = base

    @staticmethod
    def defaults(kwargs: Dict) -> Dict[str, Union[str, bool, List[str], Optional[int]]]:
        """ Return default arguments merged with user provided data. """

        _defaults = {
            'root': '.',
            'check_external_urls': False,
            'ignore_domains': [],
            'ignore_pathes': [],
            'stay_within_path': True,
            'retry': None,
            'threads': None,
        }

        return {**_defaults, **kwargs}

    # -- Root ------------------------------------------------------------------

    @property
    def masked(self) -> bool:
        """ Is Base url masked? """

        return self._is_masked

    @property
    def root(self) -> str:
        """ Getter for Document Root value. """
        return self._root

    @root.setter
    def root(self, value: Path) -> None:
        if not (self._root is None): #pylint: disable-msg=C0325
            error = "root is already set to {}"
            raise DeadlinksSettingsRoot(error.format(self._root))

        if value is None:
            error = "For URL<internal> checks, Document Root is required."
            raise DeadlinksSettingsRoot(error)

        if not isinstance(value, Path):
            raise DeadlinksSettingsRoot("Document Root required to be Path-type.")

        if not value.is_dir():
            error = "Document Root ({}) not found."
            raise DeadlinksSettingsRoot(error.format(value.resolve()))

        self._root = value

    # -- Base URL --------------------------------------------------------------
    @property
    def base(self) -> BaseURL:
        """ Getter for BaseURL ."""
        return self._base

    @base.setter
    def base(self, value: BaseURL) -> None:
        if not (self._base is None): #pylint: disable-msg=C0325
            error = "BaseUrl is already set to {}"
            raise DeadlinksSettingsBase(error.format(self._base))

        if not value.is_valid():
            raise DeadlinksSettingsBase("URL {} is not valid".format(value))

        self._base = value

    # -- Ignored Domains -------------------------------------------------------
    @property
    def domains(self) -> List[str]:
        """ Getter for Ignored Domains. """
        return self._domains

    @domains.setter
    def domains(self, values: List[str]) -> None:
        if not (self._domains is None): #pylint: disable-msg=C0325
            error = "Ignored Domains is already defined"
            raise DeadlinksSettingsDomains(error)

        for domain in values:
            if not isinstance(domain, str):
                error = 'Domain "{}" is not a string'
                raise DeadlinksSettingsDomains(error.format(domain))

            # TODO - Should I add a better check for rfc 3986 ?
            if not domain:
                error = 'Empty Domain is not accepted.'
                raise DeadlinksSettingsDomains(error.format(domain))

            if len(domain) > 255:
                error = 'Domain "{}" should conform to rfc3986.'
                raise DeadlinksSettingsDomains(error.format(domain))

        self._domains = values

    # -- Ignored Pathes --------------------------------------------------------
    @property
    def pathes(self) -> List[str]:
        """ Getter for Ignored Pathes. """
        return self._pathes

    @pathes.setter
    def pathes(self, values: List[str]) -> None:
        if not (self._pathes is None): #pylint: disable-msg=C0325
            error = "Ignored Pathes is already defined."
            raise DeadlinksSettingsPathes(error)

        for path in values:
            if not isinstance(path, str):
                error = 'Path "{}" is not a string.'
                raise DeadlinksSettingsPathes(error.format(path))

            if not path:
                error = 'Empty Path is not accepted.'
                raise DeadlinksSettingsPathes(error.format(path))

        self._pathes = list(set(values)) # only uniq values

    # -- Stay with in path setting ---------------------------------------------
    @property
    def stay_within_path(self) -> bool:
        """ Default value of stay within path setting. """
        return self._stay_within_path

    @stay_within_path.setter
    def stay_within_path(self, value: bool) -> None:
        """ Getter for stay_within_path setting. """
        if not (self._stay_within_path is None): #pylint: disable-msg=C0325
            error = "Stay within path is already defined."
            raise DeadlinksSettingsPath(error)

        if not isinstance(value, bool):
            error = 'stay_within_path can\'t be anything but bool.'
            raise DeadlinksSettingsPath(error)

        self._stay_within_path = value

    # -- Retries ---------------------------------------------------------------

    """
    Retry is additional requests we going to send to url in case if it responds
    with 502,503,504. Delay between requests exponential.

        Value | Delay (seconds)
        1     | 1
        2     | 2
        3     | 4
        4     | 8
        5     | 16
        6     | 32
        7     | 64
        8     | 108
        9     | 216
        10    | 432

       Beware of setting values above 5.
    """

    @property
    def retry(self) -> int:
        """ Getter for retry information. """
        return self._retry

    @retry.setter
    def retry(self, value: Optional[int]) -> None:
        if not (self._retry is None): #pylint: disable-msg=C0325
            raise DeadlinksSettingsChange("Change not allowed")

        # retry validation
        if value is None:
            self._retry = 0
            return

        if isinstance(value, int):
            if 0 <= value <= 10:
                self._retry = value
                return

            error = 'Setting "retry" value out of the allowed range (0...10).'
            raise DeadlinksSettingsRetry(error)

        raise DeadlinksSettingsRetry('Setting "retry" is not a number')

    # -- External -------------------------------------------------------------

    """
    Should crawler check external urls existence?
        True  - Check existence of external urls
        False - Ignore external urls while indexing website
    """

    @property
    def external(self) -> bool:
        """ Getter for External Indexation State. """
        return self._external

    @external.setter
    def external(self, value: bool) -> None:
        if not (self._external is None): #pylint: disable-msg=C0325
            raise DeadlinksSettingsChange("Change not allowed")

        self._external = bool(value)

    # -- Threads -------------------------------------------------------------

    """
    Concurrent execution of indexation. "Off" by default.
    """

    @property
    def threads(self) -> int:
        """ Getter for number of threads to run """
        return int(self._threads)

    @threads.setter
    def threads(self, value: Optional[int]) -> None:
        if not (self._threads is None): #pylint: disable-msg=C0325
            raise DeadlinksSettingsChange("Change not allowed")

        if value is None:
            self._threads = 1
            return

        if isinstance(value, int):
            if 1 <= value <= 10:
                self._threads = value
                return

            error = 'Setting "threads" value out of the allowed range (1...10).'
            raise DeadlinksSettingsThreads(error)

        raise DeadlinksSettingsThreads('Setting "threads" is not a number')



# Copyright 2019 Oleg Butuzov. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
deadlinks.serving.options
~~~~~~~~~~~~~~~~~~~~~~~~~

Default options to be consumed by click if `internal` domain passed.

:copyright: (c) 2019 by Oleg Butuzov.
:license:   Apache2, see LICENSE for more details.
"""

# -- Imports -------------------------------------------------------------------

from typing import List

from click import Path

from ..clicker import OptionRaw, Options

# -- Code ----------------------------------------------------------------------


def root_option_modifier(options: Options) -> Options:
    """ this modifier will be triggered if root option passed. """

    if not options["root"]:
        return options

    if options.get("retry", 0) == 0:
        options["retry"] = 3

    if options.get("threads", 1) == 1:
        options["threads"] = 10

    return options


default_options = [] # type: List[OptionRaw]
# root, in case of URL<internal>, defaults to . (pwd or current directory)
default_options.append((
    ('root', '-R', '--root'),
    {
        'help': 'Path to the documentation\'s Document Root',
        'metavar': '',
        'type': Path(),
        'modifier': [root_option_modifier]
    },
))



# Copyright 2019 Oleg Butuzov. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
deadlinks.link
~~~~~~~~~~~~~~

Link representation based on URL object.

:copyright: (c) 2019 by Oleg Butuzov.
:license:   Apache2, see LICENSE for more details.
"""

# -- Imports -------------------------------------------------------------------

from typing import Union

from .url import URL


class Link(URL):
    """
    Link URL with benefits

    Real reason why this exists, it's impossible situation on using class as
    type definition, in one (two) of the methods.

    In Python 3.5 its OK to use quoted class name.
    In Python 3.6 we can import annotations from __future__
    What we can't do is to try catch, python version handling or
    `__future__ import` as this is compile time operation. Therefore,
    lets use Link.
    """

    def is_external(self, url: Union[URL, str]) -> bool:
        """ Check if url is external to Link object. """

        if isinstance(url, str):
            url = URL(url)
        elif not isinstance(url, URL):
            raise TypeError("url of type {}".format(type(url)))

        base_scheme, this_scheme = url.scheme, self.scheme
        base, this = url.domain, self.domain

        # we assuming that www.domain.com and domain.com are same
        pattern = "www."
        base = base[len(pattern):] if base.startswith(pattern) else base
        this = this[len(pattern):] if this.startswith(pattern) else this

        # we assume that http://domain.com and http://domain.com:80/ are same
        if this_scheme == "http" and base_scheme == "http":
            pattern = ":80"
            base = base[:-(len(pattern))] if base.endswith(pattern) else base
            this = this[:-(len(pattern))] if this.endswith(pattern) else this
        elif this_scheme == "https" and base_scheme == "https":
            pattern = ":443"
            base = base[:-(len(pattern))] if base.endswith(pattern) else base
            this = this[:-(len(pattern))] if this.endswith(pattern) else this

        return base != this

    def __eq__(self, other: object) -> bool:
        """ Compare two links. """

        if isinstance(other, str):
            other = Link(other)

        if not isinstance(other, URL):
            raise TypeError("`url` expected to be str or Link ")

        return self.url() == other.url()



# Copyright 2019 Oleg Butuzov. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
deadlinks.options
~~~~~~~~~~~~~~~~~

Default options to be consumed by click

:copyright: (c) 2019 by Oleg Butuzov.
:license:   Apache2, see LICENSE for more details.
"""

# -- Imports -------------------------------------------------------------------

from typing import List

from click import IntRange, Choice

from .clicker import OptionRaw

# -- Options -------------------------------------------------------------------

default_options = [] # type: List[OptionRaw]

# Index External URLs ------------------------------------------------------
default_options.append((
    ("check_external_urls", '-e', '--external'),
    {
        'default': False,
        'is_flag': True,
        'show_default': False,
        'help': 'Enables external resources check',
    },
))

# Retries ------------------------------------------------------------------
default_options.append((
    ('-r', '--retry'),
    {
        'default': 0,
        'type': IntRange(0, 10),
        'is_flag': False,
        'show_default': True,
        'metavar': '',
        'help': 'Number of Retries [0...10]',
    },
))

# Concurrent Crawlers ------------------------------------------------------
default_options.append((
    ('-n', '--threads'),
    {
        'default': 1,
        'type': IntRange(1, 10),
        'is_flag': False,
        'show_default': True,
        'metavar': '',
        'help': 'Concurrent crawlers [1...10]',
    },
))

# Ignored Domains  ---------------------------------------------------------
default_options.append((
    ('ignore_domains', '-d', '--domain'),
    {
        'multiple': True,
        'metavar': '',
        'help': 'Domain to ignore (multiple options allowed)',
    },
))

# Ignored Paths  -----------------------------------------------------------
default_options.append((
    ('ignore_pathes', '-p', '--path'),
    {
        'multiple': True,
        'metavar': '',
        'help': 'Path to ignore (multiple options allowe)',
    },
))

# Disable path limiting for a local to the domain links.
default_options.append((
    ('stay_within_path', '--full-site-check'),
    {
        'default': True,
        'is_flag': True,
        'help': 'Check full site',
    },
))

# Show selectors.
default_options.append((
    ('show', '-s', '--show'),
    {
        'default': ['failed'],
        'multiple': True,
        'type': Choice(
            ['failed', 'ok', 'ignored', 'all', 'none'],
            case_sensitive=False,
        ),
        'show_default': False,
        'metavar': '',
        'help': 'Show results [failed (default), ok, ignored, all or none]',
    },
))

default_options.append((
    ('fail_if_fails_found', '--fiff'),
    {
        'default': False,
        'is_flag': True,
        'help': 'Fail if failed URLs are found',
    },
))



# Copyright 2019 Oleg Butuzov. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
deadlinks.main
~~~~~~~~~~~~~~

Main (cli interface)

:copyright: (c) 2019 by Oleg Butuzov.
:license:   Apache2, see LICENSE for more details.
"""

# -- Imports -------------------------------------------------------------------

import click

from .settings import Settings
from .crawler import Crawler
from .exceptions import DeadlinksException

# CLI implementation related
from .clicker import (register_options, register_exports)
from .clicker import Options
from .clicker import (command, argument)

# Exporters
from .exporters import Export #pylint: disable-msg=W0611
from .exporters import exporters

# Default and Specified Options
from .options import default_options as general_options
from .serving.options import default_options as serving_options

# Version and App.
from .__version__ import __app_version__ as version
from .__version__ import __app_package__ as name

# -- Implementation ------------------------------------------------------------


@click.command(name, **command)
@click.argument('url', **argument)
@click.version_option(version, '-V', '--version', message='%(prog)s: v%(version)s', prog_name=name)
@register_exports(exporters)
@register_options("Server Settings", serving_options)
@register_options("Default Options", general_options)
@click.pass_context
def main(ctx: click.Context, url: str, **opts: Options) -> None:
    """ Check links in your (web) documentation for accessibility. """

    try:
        settings = Settings(url, **opts)
        crawler = Crawler(settings)

        driver = exporters[str(opts['export'])]

        # Instantion of the exported before starting crawling will alow us to
        # have progress report, while we crawling website.
        exporter = driver(crawler, **opts)

        # Starting crawler
        crawler.start()

        # And showing report of crawling, in 90% we not showing anything
        # un till crawling is done, so we can just output (pipe) data
        # to where user desire have results.
        exporter.report()

        if opts['fail_if_fails_found'] and len(crawler.failed) > 0:
            ctx.exit(1)

    except DeadlinksException as e:
        ctx.fail(e.__str__())


if __name__ == "__main__":
    main() # pylint: disable=no-value-for-parameter



# Copyright 2019 Oleg Butuzov. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
deadlinks.exporters.__init__
~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Export module init file.

:copyright: (c) 2019 by Oleg Butuzov.
:license:   Apache2, see LICENSE for more details.
"""

# -- Imports -------------------------------------------------------------------

from typing import (Dict, Tuple, List, Sequence, Any) #pylint: disable-msg=W0611

from threading import Thread
from os import name as os_name
from time import sleep

import click

from .export import Export
from ..crawler import Crawler
from ..clicker import OptionRaw

# -- Implementation ------------------------------------------------------------

BEFORE_BAR = '\r' if os_name == 'nt' else '\r\033[?25l'
AFTER_BAR = '\n' if os_name == 'nt' else '\033[?25h\n'


class Default(Export):

    PROGRESS = "Ready: {0:>3.0%} ({1}/{2});"
    FIN_DONE = "Links Total: {};"
    DETAILED = "Found: {0}; Not Found: {1}; Ignored: {2}; Redirects: {3}"

    params_colors = {
        'succeed': 'green',
        'failed': 'red',
        'ignored': 'yellow',
    }

    def __init__(self, crawler: Crawler, **opts: Dict) -> None:
        self._crawler = crawler
        self._opts = opts

        self._progress_msg = ""
        self._progress_trd = Thread(target=self._progress_handler, daemon=True)
        self._progress_trd.start()

    def is_colored(self) -> bool:
        """ provides check about colored output """
        return not self._opts.get('no_colors', False)

    def info(self) -> str:

        base = self._crawler.settings.base
        baseurl = base if not self._crawler.settings.masked else \
                    base.url().replace(base.domain, 'internal')

        message = "URL=<{}>; External Checks={}; Threads={}; Retry={}".format(
            baseurl,
            "On" if self._crawler.settings.external else "Off",
            self._crawler.settings.threads,
            self._crawler.settings.retry,
        )

        return message

    @staticmethod
    def options() -> Tuple[str, List[OptionRaw]]:

        options = [] # type: List[OptionRaw]

        # Default export
        options.append((
            ('--export', ),
            {
                'default': 'default',
                'hidden': True,
                'multiple': False,
                'type': click.Choice(['default'], case_sensitive=False),
                'help': 'Export type',
            },
        ))

        options.append((
            ('--no-colors', ),
            {
                'default': False,
                'is_flag': True,
                'help': 'Color output of `default` export',
            },
        ))

        options.append((
            ('--no-progress', ),
            {
                'default': False,
                'is_flag': True,
                'help': 'Disable Proogresion output',
            },
        ))

        return ("Exporter (default)", options)

    def _progress_handler(self) -> None:
        """ progress handler desides states regarding crawler """

        # user disables progression reports.
        if self._opts['no_progress']:
            return

        while not (self._crawler.crawled or self._crawler.terminated):
            if self._crawler.crawling:
                click.echo(BEFORE_BAR, nl=False)

                unstyled_text_len = len(click.unstyle(self._progress_msg))
                click.echo(' ' * unstyled_text_len, nl=False)
                click.echo(BEFORE_BAR, nl=False)

                self._progress_msg = self._get_progress()
                click.echo(self._progress_msg, color=self.is_colored(), nl=False)
                sleep(0.1)

    def _get_progress(self) -> str:
        """ get progress report from crawler"""

        # ready / total
        stats = dict() # type: Dict[str, str]
        total = 0 # type: int
        for k, v in self._crawler.stats.items():
            total += v
            stats[str(k)] = str(v)

        total_links = len(self._crawler.index)

        progress = self.PROGRESS.format(total / total_links, total, total_links)

        if self._crawler.crawled:
            progress = self.FIN_DONE.format(total_links)

        detailed = self.DETAILED.format(
            click.style(stats['Status.FOUND'], fg='green'),
            click.style(stats['Status.NOT_FOUND'], fg='red'),
            click.style(stats['Status.IGNORED'], fg='yellow'),
            click.style(stats['Status.REDIRECTION'], fg='blue'),
        )

        return progress + ' ' + detailed

    def _generate(self, key: str) -> str:
        """ generate a report about urls """

        param = 'succeed' if key == 'ok' else key

        links = self._crawler.__getattribute__(param)
        if not links:
            return ""

        param_color = click.style(param, fg=self.params_colors[param])

        return '\n'.join(map(lambda x: "[ {} ] {}".format(param_color, x), links))

    def report(self) -> None:

        # progress wasn't disabled, so we need to cleanup a bit.
        clear_line_width = len(click.unstyle(self._progress_msg))

        if self._crawler.terminated:
            # we adding additional 10 spaces to clear ^C we got on the
            # terminal if ^C pressed.
            clear_line_width += 10

        if not self._opts['no_progress']:
            click.echo(BEFORE_BAR, nl=False)
            click.echo(' ' * clear_line_width, nl=False)
            click.echo(BEFORE_BAR, nl=False)

        info = self.info() # type: str
        stat = self._get_progress() # type: str

        split_line_len = max(len(click.unstyle(info)), len(click.unstyle(stat)))

        if self._crawler.terminated:
            term_msg = "Results can be inconsistent, as execution was terminated."
            pref_len = (split_line_len - len(term_msg)) // 2
            click.echo((" " * split_line_len))
            click.echo((" "*pref_len) + term_msg)
            click.echo((" "*pref_len) + ("^" * len(term_msg)))

        click.echo("=" * split_line_len)
        click.echo(info, color=self.is_colored())
        click.echo("=" * split_line_len)

        click.echo(stat, color=self.is_colored())
        click.echo(("-"*split_line_len) + "\033[?25h")

        # show some url report(s)
        show = list(self._opts.get('show', [])) # type: Sequence[str]

        if 'none' in show:
            return

        if 'all' in show:
            show = ['ok', 'failed', 'ignored']

        for report in show:
            OUTPUT_REPORT = self._generate(report)

            if len(OUTPUT_REPORT):
                click.echo(OUTPUT_REPORT, color=self.is_colored())



# Copyright 2019 Oleg Butuzov. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
deadlinks.clicker
~~~~~~~~~~~~~~~~~

Functions related to click package and cli implementation.

:copyright: (c) 2019 by Oleg Butuzov.
:license:   Apache2, see LICENSE for more details.
"""

# -- Imports -------------------------------------------------------------------

from typing import (Dict, List, Tuple, Union, Any, Callable, Sequence)
from textwrap import dedent
from collections import OrderedDict

from click import Option, Argument
from click import Command, Context
from click import HelpFormatter as Formatter

from .link import Link
from .__version__ import __app_package__ as app

# -- Typing Decorators -----~---------------------------------------------------
OptionsValues = Union[str, bool, int, List[str], List[Callable]]
Options = Dict[str, OptionsValues]
OptionModifier = Callable
OptionRaw = Tuple[Sequence[str], Dict[str, Any]]
OptionsList = List[OptionRaw]
HelpDefList = List[Tuple[str, str]]

# -- Options Decorators --------------------------------------------------------
#   Its better to make own wrappet than add number of options directly to
#   main function each time.


class OrderedDefaultDict(OrderedDict):
    factory = list

    def __missing__(self, key: str) -> List:
        self[key] = value = self.factory() # type: ignore
        return value


default_attributes = {
    '__click_params_groups__': OrderedDefaultDict,
    '__click_params__': list,
    '__click_params_modify__': OrderedDefaultDict,
}


def modify(f: Any, group: str, options: OptionsList) -> Any:

    for attr, factory in default_attributes.items():
        if not hasattr(f, attr):
            f.__dict__[attr] = factory()

    for option in reversed(options):
        # exported value is variable that we getting in the function.
        # params is a list of options
        # key=>value
        params, attrs = option

        OptionClass = attrs.pop('cls', Option)
        Modifiers = attrs.pop('modifier', [])

        _option = OptionClass(params, **attrs)

        # adding option and group
        f.__click_params__.append(_option)
        f.__click_params_groups__[group].append(_option.name)
        if Modifiers:
            f.__click_params_modify__[_option.name] += Modifiers

    return f


def register_exports(exporters: Dict[str, Any]) -> Callable:

    def decorator(f: Any) -> Any:

        for exporter in exporters.values():
            group, options = exporter.options()
            f = modify(f, group, options)

        return f

    return decorator


def register_options(group: str, options: OptionsList) -> Callable:
    """ Register Multiple Options in one set  """

    def decorator(f: Any) -> Any:
        """ Final decoration for main """

        return modify(f, group, options)

    return decorator


# -- CLI wrapper ---------------------------------------------------------------
#   So we can have nice looking options groups.


class Clicker(Command):
    """ For a better --help function """

    EXAMPLES = dedent(
        """\
            // Check links (including external) at http://gobyexample.com/ in 10 threads,
            // but not ones that leading to domains play.golang.org or github.com
            deadlinks gobyexample.com -n 10 -e -d play.golang.org -d github.com

            // Limiting check only to links found within /docs path.
            deadlinks http://localhost:1313/docs

            // Running checks for all local links that belong to a domain.
            deadlinks http://localhost:1313/docs/ -n 10 --full-site-check
        """)

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """ Getting Options groups """

        if hasattr(kwargs['callback'], '__click_params_groups__'):
            self._groups = kwargs['callback'].__click_params_groups__
            self._groups['Other'] = []

        if hasattr(kwargs['callback'], '__click_params_modify__'):
            self._modify = kwargs['callback'].__click_params_modify__

        super().__init__(*args, **kwargs)

    def format_help(self, ctx: Context, formatter: Formatter) -> None:
        """ Writes the help into the formatter if it exists. """

        # Checking links from web resource for dead/alive status.
        self.format_help_text(ctx, formatter)
        self.format_examples(ctx, formatter)
        self.format_options(ctx, formatter)
        self.format_epilog(ctx, formatter)

    def make_context(self, info_name, args, parent=None, **extra) -> Context: # type: ignore
        """ Hooking into context in oder to make pro_name for unnamed call. """

        extra['help_option_names'] = ['--help', '-h']

        return super().make_context(app, args, parent, **extra)

    def format_options(self, ctx: Context, formatter: Formatter) -> None:
        """ Group options together. """

        if not hasattr(self, '_groups'):
            return super().format_options(ctx, formatter)

        # building reverse look up
        lookup = {}
        for group, items in self._groups.items():
            lookup.update({item: group for item in items})

        groups = {group: [] for group in self._groups} #type: Dict[str, HelpDefList]

        for param in self.get_params(ctx):
            rv = param.get_help_record(ctx)
            if rv is None:
                continue

            current_group = lookup.get(param.name, "Other")
            groups[current_group].append(rv)

        for group_name, options in groups.items():
            if not options:
                continue

            with formatter.section(group_name):
                formatter.write_dl(options)

    def format_help_text(self, ctx: Context, formatter: Formatter) -> None:
        """ Writes the help text to the formatter if it exists. """

        formatter.write_paragraph()
        formatter.write_text(self.help)

    def format_examples(
            self, ctx: Context, formatter: Formatter) -> None: # pylint: disable-msg=R0201
        """ Writes the examples into the formatter. """

        with formatter.section('Usage Examples'):
            formatter.write("\n")
            for line in self.EXAMPLES.split("\n"):
                formatter.write("  {}\n".format(line))

    def modify(self, ctx: Context) -> None:
        """ Modifing params based on context. """
        if not self._modify:
            return

        for key, modifiers in self._modify.items():
            if ctx.params.get(key, None):
                for modify in modifiers:
                    ctx.params = modify(ctx.params)

    def invoke(self, ctx: Context) -> None:
        """ Hooking into invoke in order to modify params. """
        self.modify(ctx)
        super().invoke(ctx)


# -- CLI wrapper ---------------------------------------------------------------
#   So we can have nice looking options groups.

command = dict({
    "cls": Clicker,
    "context_settings": {
        "ignore_unknown_options": False
    },
}) # type: Dict[str, Any]


def validate_url(ctx: Context, param: Argument, value: str) -> str:
    """ if received url with no scheme will try to fix it """

    url = Link(value)
    if not url.is_valid() and not url.scheme:
        return "http://{}".format(url)

    return value


argument = dict({
    'nargs': 1,
    'required': True,
    'callback': validate_url,
    'metavar': '<URL>',
}) # type: Dict[str, Any]



# Copyright 2019 Oleg Butuzov. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
deadlinks.baseurl
~~~~~~~~~~~~~~~~~

Starting Point URL or BaseURL

:copyright: (c) 2019 by Oleg Butuzov.
:license:   Apache2, see LICENSE for more details.
"""

from .link import Link


class BaseURL(Link):
    """ BaseURL propose to allow subpath check for provided link

        Having this class allow us to pass responsibility for
        "checking" of within class to it, instead adding this functionality
        to crawler or Link.
    """

    def __init__(self, url: str) -> None:
        super().__init__(url)
        self._path = self.get_base_path()

    def get_base_path(self) -> str:
        file = self.path.split("/")[-1]
        if "." in file:
            return self.path[:(len(self.path) - len(file))]

        return self.path

    def within(self, link: Link) -> bool:
        """ determite is `link` argument stays within path of baseurl """

        if link.is_external(self):
            return False

        return link.path.startswith(self._path.rstrip("/"))


if __name__ == "__main__":

    b1 = BaseURL("http://google.com/somepath/index.html")
    print(b1.within(Link("http://google.com/somepath/specs.html")))

    b2 = BaseURL("http://google.com/index.html")
    print(b2.within(Link("http://google.com/somepath/specs.html")))

    b3 = BaseURL("http://google.com/")
    print(b3.within(Link("http://google.com/somepath/specs.html")))



# Copyright 2019 Oleg Butuzov. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
deadlinks.serving.handler
~~~~~~~~~~~~~~~~~~~~~~~~~

Handles requests responses via simple webserver.

:copyright: (c) 2019 by Oleg Butuzov.
:license:   Apache2, see LICENSE for more details.
"""

# -- Imports -------------------------------------------------------------------

from typing import Any

try:
    from http.server import BaseHTTPRequestHandler
except ModuleNotFoundError:
    from BaseHTTPServer import BaseHTTPRequestHandler # type: ignore

from .router import Router

# -- Implementation ------------------------------------------------------------


class Handler(BaseHTTPRequestHandler):

    def __init__(self, router: Router, *args: Any, **kwargs: Any) -> None:
        """ Defined logic for responding on static files requests. """
        self._router = router

        super().__init__(*args, **kwargs)

    def log_message(self, *args: Any) -> None:
        """ Ignoring logging """

    def do_GET(self) -> None:
        """ SuperSimple GET handler.

        Responds with (addording router results for requested path):
            Code 200 -> Content
            Code 301 -> New Location
            Code 404 -> Nothing (None)
        """

        code, response = self._router(self.path)

        if code == 301:
            self.send_response(code)
            self.send_header('Location', response)
            self.end_headers()
            return

        if code == 404:
            self.send_response(code)
            self.end_headers()
            return

        try:

            with open(response, "r") as f:

                try:
                    self.send_response(200)
                    self.end_headers()
                    self.wfile.write(f.read().encode())
                except UnicodeDecodeError:
                    pass

        except FileNotFoundError:
            self.send_response(404)
            self.end_headers()
            return



# Copyright 2019 Oleg Butuzov. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
deadlinks.main
~~~~~~~~~~~~~~

Main (cli interface)

:copyright: (c) 2019 by Oleg Butuzov.
:license:   Apache2, see LICENSE for more details.
"""

# -- Imports -------------------------------------------------------------------

from typing import (Union, Optional)

from functools import partial

try:
    from socketserver import ThreadingMixIn
    from http.server import HTTPServer
except ModuleNotFoundError:
    from SocketServer import ThreadingMixIn # type: ignore
    from BaseHTTPServer import HTTPServer # type: ignore

from socket import (socket, SOCK_STREAM, AF_INET)
from threading import Thread
from pathlib import Path

from .handler import Handler
from .router import Router

# -- Implementation ------------------------------------------------------------


class ThreadedHTTPServer(ThreadingMixIn, HTTPServer):
    """Handle requests in a separate thread."""
    daemon_threads = True


class SimpleServer:

    def __init__(self, web_root: Union[str, Path], web_path: Optional[str]) -> None:
        """ Starts simple webserver and handles requests to local folder. """
        self.web_path = "/" if not web_path else web_path
        if not self.web_path.startswith("/"):
            self.web_path = "/" + self.web_path

        if not isinstance(web_root, Path):
            web_root = Path(web_root)

        self.router = Router(web_root.resolve(), self.web_path)

        _socket = socket(AF_INET, type=SOCK_STREAM)
        _socket.bind(('localhost', 0))
        self._sa = _socket.getsockname()
        _socket.close()

        self.start()

    def start(self) -> None:

        # implement correct type annotation, when change
        # https://github.com/python/mypy/issues/1484
        self._handler = partial(Handler, self.router)
        self._server = ThreadedHTTPServer(self._sa, self._handler) # type: ignore
        server_thread = Thread(target=self._server.serve_forever, daemon=True)
        server_thread.start()

    def __str__(self) -> str:
        """ Instance as browsable URL. """

        return self.url()

    def url(self) -> str:
        """ Return URL of running server (including path). """
        return "http://{}:{}{}".format(self._sa[0], self._sa[1], self.web_path)



"""
deadlinks
~~~~~~~~~
deadlinks checker for your static website. It's better keep house clean, right?
"""

from typing import (Dict, Tuple, Optional, List) #pylint: disable-msg=W0611

from pathlib import Path
from re import match
import os
from setuptools import find_packages, setup
from setuptools.command.build_py import build_py as _build_py

from utils.setup import (read_data, require, readme)
try:
    from utils.brew import build_formula
except BaseException:
    build_formula = lambda a, b: "Brew Build not implemented"

# ------------------------------------------------------------------------------

# ~~ Version Releases / Start ~~

# PyPi: only releases (x.y.z)
#
#
data = read_data()

branch = os.environ.get('DEADLINKS_BRANCH', None)
commit = os.environ.get('DEADLINKS_COMMIT', None)
tagged = os.environ.get('DEADLINKS_TAGGED', None)

VERSION = r'^\d{1,}.\d{1,}.\d{1,}$' # type: str

if tagged and not match(VERSION, tagged) and branch and commit:
    dev_version_file = Path(__file__).parent / "deadlinks" / "__develop__.py"
    dev_version_str = ".{}.{}".format(branch, commit).rstrip("+")
    with open(str(dev_version_file), "w") as f:
        print("version = '{}'".format(dev_version_str), file=f)
    data['app_version'] += dev_version_str

# -- Version Releases / End ~~


class BrewFormulaBuilder(_build_py):
    """ Run external formula builder """

    def run(self) -> None:
        data = read_data()
        requirements = require("install") + require("brew")

        with open("{}.rb".format(data['app_package']), 'w') as f:
            print(build_formula(app=data, requirements=requirements), file=f)


# -- Setup ---------------------------------------------------------------------

if __name__ == "__main__":
    setup(
        name=data['app_package'],
        version=data['app_version'],
        description=data['description'],
        long_description=readme(),
        long_description_content_type="text/markdown",
        keywords=["documentation", "website", "spider", "crawler", "link-checker"],
        author=data['author_name'],
        author_email=data['author_mail'],
        project_urls={
            "GitHub: repo": "https://github.com/butuzov/deadlinks",
            "Bugtracker": "https://github.com/butuzov/deadlinks/issues",
            "Documentation": "http://deadlinks.readthedocs.io/",
            "Documentation (latest)": "https://deadlinks.readthedocs.io/en/latest/",
            "Dockerized": "https://hub.docker.com/repository/docker/butuzov/deadlinks/",
        },
        packages=find_packages(exclude=["utils*", "unittests*"]),
        install_requires=require("install"),
        entry_points='''
            [console_scripts]
            deadlinks=deadlinks.__main__:main
        ''',
        zip_safe=False,
        cmdclass={'brew_formula_create': BrewFormulaBuilder},
        python_requires='>=3.5',
        url=data['app_website'],
        license=data['app_license'],
        platforms=['MacOS', 'Posix', 'Unix'],
        classifiers=[
            # Env
            "Environment :: Console",

            # Status
            "Development Status :: 2 - Pre-Alpha",

            # Audience
            "Intended Audience :: Developers",
            "Intended Audience :: System Administrators",

            # Topic
            "Topic :: Utilities",
            "Topic :: Documentation",

            # Audience and Topic
            "Intended Audience :: Developers",
            "Intended Audience :: System Administrators",

            # Python version
            "Programming Language :: Python :: 3 :: Only",
            "Programming Language :: Python :: 3.5",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",

            # License
            "License :: OSI Approved :: Apache Software License",

            # Operation System
            "Operating System :: MacOS",
            "Operating System :: POSIX",
            "Operating System :: Unix",

            # Language
            "Natural Language :: English",
        ],
    )



# Copyright 2019 Oleg Butuzov. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
deadlinks.serving.router
~~~~~~~~~~~~~~~~~~~~~~~~

Implements simple file system lookups for web requests on generated docs.

:copyright: (c) 2019 by Oleg Butuzov.
:license:   Apache2, see LICENSE for more details.
"""

# -- Imports -------------------------------------------------------------------

from typing import (Union, Dict, Optional, Tuple)

from pathlib import Path
from collections import OrderedDict

from ..exceptions import DeadlinksSettingsRoot

# -- Implementation ------------------------------------------------------------

root_web_files = [
    '/robots.txt',
    '/favicon.ico',
    '/sitemap.xml',
]


class Router():
    """ Router for the static site  """

    def __init__(self, siteroot: Union[Path, str], sitepath: str) -> None:
        """ Transform startup params and load redirections. """

        if isinstance(siteroot, str):
            siteroot = Path(siteroot)

        self._siteroot = siteroot
        self._sitepath = sitepath

        # handle not found error.
        if not self._siteroot.exists():
            raise DeadlinksSettingsRoot("DocumentRoot not found")

        if not self._siteroot.is_dir():
            raise DeadlinksSettingsRoot("This is not a directory")

        # redirects
        self._redirects = OrderedDict() # type: Dict[str, str]
        self.load_redirects()

    def load_redirects(self) -> None:
        """  Supports only _redirects - docs.netlify.com/routing/redirects . """

        self._redirects_file = self._siteroot / "_redirects"

        if not self._redirects_file.exists():
            return

        with open(str(self._redirects_file)) as file:
            for line in file:
                if not line or line.strip() == "" or line[0] == "#":
                    continue

                p = line.split()
                if len(p) < 2:
                    continue

                self._redirects[p[0]] = p[1]

    def __call__(self, request_url: str) -> Tuple[int, Optional[str]]:
        """ Implements router logic. """

        # request_url is redirect
        if request_url in self._redirects:
            return (301, self._redirects[request_url])

        # requests for favicon, robots.txt and sitemap.xml
        if request_url in root_web_files:
            request_file = self._siteroot / request_url.lstrip("/")
            if request_file.is_file():
                return (200, str(request_file.resolve()))

            return (404, None)

        if request_url.startswith(self._sitepath):
            request_url = request_url[len(self._sitepath):]

        request_file = self._siteroot / request_url.lstrip("/")

        try_files = []

        if request_file.is_dir():
            try_files.append(request_file / "index.html")
            try_files.append(request_file / "index.htm")
        else:
            _path = str(request_file)
            try_files.append(Path(_path + ".html"))
            try_files.append(Path(_path + ".htm"))
            try_files.append(request_file)

        for file in try_files:
            if file.is_file():
                return (200, str(file))

        return (404, None)



# Copyright 2019 Oleg Butuzov. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
deadlinks.exporters.__init__
~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Export module init file.

:copyright: (c) 2019 by Oleg Butuzov.
:license:   Apache2, see LICENSE for more details.
"""

# -- Imports -------------------------------------------------------------------

from typing import Dict #pylint: disable-msg=W0611

from .export import Export
from .default import Default

# -- Exports -------------------------------------------------------------------

__all__ = [
    'Default',
    'Export',
]

exporters = { # type Dict[str, Export]
    'default': Default,
}
